import os
import logging
from datetime import datetime as dt
logger = logging.getLogger(__name__)


def _check_cache(cache):
    """Check the cache folder directory exists
    Parameters
    ----------
    cache : str
        Folder path location to be used as local cache.
    Returns
    -------
    str
        Existing (or updated, if error occurs) cache directory path.
    """
    try:
        assert(isinstance(cache, str))
    except AssertionError as e:
        logger.exception("Cache input must be a string")
        raise e

    if not os.path.isdir(cache):
        cache = os.path.join(os.getcwd(), "cache_nemed")
        if os.path.isdir(cache):
            logger.warning(f"Input cache path is invalid or not found. Using existing cache directory at {cache}")
        else:
            os.mkdir(cache)
            logger.warning(f"Input cache path is invalid or not found. Creating new cache directory at {cache}")
    return cache


def _validate_variable_type(var, vartype, inputname):
    try:
        assert isinstance(var, vartype), f"`{inputname}` cannot be {var}. Accepted type: {vartype}"
    except AssertionError as e:
        logger.exception(e)
        raise e


def _validate_and_convert_date(date_string, inputname, date_format="%Y/%m/%d %H:%M"):
    try:
        date_obj = dt.strptime(date_string, date_format)
        return date_obj
    except ValueError:
        e = (f"Invalid date string was passed {date_string} as {inputname}. Required date format: {date_format}")
        logger.exception(e)


""" Modifies functionality from nemosis to extract emissions factors"""
import logging
import os as _os
import glob as _glob
from nemosis import defaults as _defaults
from nemosis import processing_info_maps as _processing_info_maps
from nemosis import data_fetch_methods
from nemosis.data_fetch_methods import _create_filename, _download_data, _get_read_function, \
    _determine_columns_and_read_csv, _perform_column_selection, _log_file_creation_message, _write_to_format

logger = logging.getLogger(__name__)

def overwrite_nemosis_defaults():
    """Overwrites nemosis defaults with table properties for GENUNITS from AEMO NEMWEB.
    """
    _defaults.table_types.update({"GENUNITS": "MMS", "DUALLOC": "MMS"})
    _defaults.names.update({"GENUNITS": "PUBLIC_DVD_GENUNITS", "DUALLOC": "PUBLIC_DVD_DUALLOC"})
    _defaults.data_url.update({"GENUNITS": "aemo_data_url", "DUALLOC": "aemo_data_url"})
    _defaults.table_columns.update({"GENUNITS": ['GENSETID', 'LASTCHANGED', 'VOLTLEVEL', 'REGISTEREDCAPACITY', \
        'DISPATCHTYPE', 'STARTTYPE', 'MAXCAPACITY', 'GENSETTYPE', 'CO2E_EMISSIONS_FACTOR', 'CO2E_ENERGY_SOURCE', \
        'CO2E_DATA_SOURCE'],
        "DUALLOC": ["EFFECTIVEDATE", "VERSIONNO", "DUID", "GENSETID", "LASTCHANGED"]})


def mod_dynamic_data_fetch_loop(
    start_search,
    start_time,
    end_time,
    table_name,
    raw_data_location,
    select_columns,
    date_filter,
    fformat="feather",
    keep_csv=True,
    caching_mode=False,
    rebuild=False,
    write_kwargs={},
):
    """Modified function from nemosis.data_fetch_methods
    """
    data_tables = []

    table_type = _defaults.table_types[table_name]
    date_gen = _processing_info_maps.date_gen[table_type](start_search, end_time)

    for year, month, day, index in date_gen:
        filename_stub, full_filename, path_and_name = _create_filename(
            table_name, table_type, raw_data_location, fformat, day, month, year, index
        )

        if not (
            _glob.glob(full_filename) or _glob.glob(path_and_name + ".[cC][sS][vV]")
        ) or (not _glob.glob(path_and_name + ".[cC][sS][vV]") and rebuild):
            _download_data(
                table_name,
                table_type,
                filename_stub,
                day,
                month,
                year,
                index,
                raw_data_location,
            )

        if _glob.glob(full_filename) and fformat != "csv" and not rebuild:
            if not caching_mode:
                data = _get_read_function(fformat, table_type, day)(full_filename)
                data.insert(0, 'file_month', month)
                data.insert(0, 'file_year', year)
            else:
                data = None
                logger.info(
                    f"Cache for {table_name} in date range already compiled in"
                    + f" {raw_data_location}."
                )

        elif _glob.glob(path_and_name + ".[cC][sS][vV]"):

            if select_columns != "all":
                read_all_columns = False
            else:
                read_all_columns = True

            if not caching_mode:
                dtypes = "str"
            else:
                dtypes = "all"

            csv_path_and_name = _glob.glob(path_and_name + ".[cC][sS][vV]")[0]

            csv_read_function = _get_read_function(
                fformat="csv", table_type=table_type, day=day
            )
            data = _determine_columns_and_read_csv(
                table_name,
                csv_path_and_name,
                csv_read_function,
                read_all_columns=read_all_columns,
                dtypes=dtypes,
            )
            data.insert(0, 'file_month', month)
            data.insert(0, 'file_year', year)
            if caching_mode:
                data = _perform_column_selection(data, select_columns, full_filename)
                data.insert(0, 'file_month', month)
                data.insert(0, 'file_year', year)
            if data is not None and fformat != "csv":
                _log_file_creation_message(fformat, table_name, year, month, day, index)
                data.drop(['file_month','file_year'], axis=1,inplace=True)
                _write_to_format(data, fformat, full_filename, write_kwargs)

            if not keep_csv:
                _os.remove(_glob.glob(path_and_name + ".[cC][sS][vV]")[0])
        else:
            data = None

        if not caching_mode and data is not None:

            if date_filter is not None:
                data = date_filter(data, start_time, end_time)

            data = _perform_column_selection(data, select_columns, full_filename)
            data.insert(0, 'file_month', month)
            data.insert(0, 'file_year', year)
            data_tables.append(data)
        elif not caching_mode:
            logger.warning(f"Loading data from {full_filename} failed.") # logger.
        
    return data_tables


# def overwrite_nemosis_data_fetch():
#     """Overwrites nemosis dyanmic_data_fetch_loop with added file year-month properties inserted into extracted data
#     table from AEMO NEMWEB.
#     """
#     data_fetch_methods._dynamic_data_fetch_loop = mod_dynamic_data_fetch_loop




""" Process functions for calculations based on downloaded data """
from datetime import datetime as dt, timedelta
import pandas as pd
import numpy as np
import logging
import os
from .downloader import download_cdeii_table, download_unit_dispatch, download_pricesetter_files, download_generators_info, \
    download_duid_auxload, download_plant_emissions_factors, read_plant_auxload_csv, download_genset_map, download_dudetailsummary
from .helper_functions import helpers as hp
from .defaults import CO2E_DATA_SOURCE_YEARMAP

DISP_INT_LENGTH = 5
logger = logging.getLogger(__name__)


def get_total_emissions_by_DI_DUID(start_time, end_time, cache, filter_regions=None, generation_sent_out=True, \
                                   assume_energy_ramp=True, dropna_co2factors=True, return_all=False):
    """Retrieve the total emissions for each generation unit per dispatch interval.

    Parameters
    ----------
    start_time : str
        Start Time Period in format 'yyyy/mm/dd HH:MM'
    end_time : str
        End Time Period in format 'yyyy/mm/dd HH:MM'
    cache : str
        Raw data location in local directory
    filter_regions : list(str)
        NEM regions to filter for while retrieving the data, as a list, by default None to collect all region data
    generation_sent_out : bool
        Considers 'sent_out' generation (auxilary loads) as opposed to 'as generated' in calculations, by default True
    assume_energy_ramp : bool
        Uses a linear ramp between dispatch scada points as opposed to a stepped function, by default True
    dropna_co2factors : bool
        Removes data (generation) entries which do not have a CO2E_EMISSIONS_FACTOR mapped to them, by default True
    return_all : bool
        Returns the entire table will all columns as opposed to tidied up table, by default False

    Returns
    -------
    pandas.DataFrame
        Data is returned as formatted if `return_all` = False, `generation_sent_out` = True:

        =========================  ========  ===================================================================================================================
        Columns:                   Type:     Description:
        DUID                       str       Generator Identifier.
        Time                       datetime  Timestamp for end of interval.
        Region                     str       The NEM region corresponding to data.
        Plant_Emissions_Intensity  float     The CO2_EMISSIONS_FACTOR [tCO2-e/MWh] corresponding to DUID.
        Energy                     float     The energy [MWh] (as generated) calculated as step or ramp depending on `assume_energy_ramp`.
        PCT_AUXILIARY_LOAD         int       The percentage of auxiliary load corresponding to DUID.
        Energy_SO                  float     The energy [MWh] (sent out) calculated based on Energy and PCT_AUXILIARY_LOAD
        Total_Emissions            float     The emissions [tCO2-e] for the DUID and Time based on Energy_SO and Plant_Emissions_Intensity
        =========================  ========  ===================================================================================================================

    """
    # Check if cache is an existing directory
    hp._check_cache(cache)

    # Adjust to also collect prior DI to calculate energy ramp 
    actual_stime = dt.strptime(start_time, "%Y/%m/%d %H:%M")
    actual_etime = dt.strptime(end_time, "%Y/%m/%d %H:%M")
    if actual_etime < actual_stime:
        raise Exception("end_time cannot be prior start_time")

    prior_start_time = actual_stime - timedelta(minutes=DISP_INT_LENGTH)
    prior_start_time = dt.strftime(prior_start_time, "%Y/%m/%d %H:%M")

    # Segment emissions calculations into smaller chunks
    ts = _generate_timeseries_loop(prior_start_time, end_time)
    res_str = []
    for sdate, edate, st, et in zip(ts['start'], ts['end'], ts['s_str'], ts['e_str']):
        logger.info(f"Processing total emissions from {st} to {et}")
        df = _total_emissions_process(sdate, edate, cache, filter_regions, generation_sent_out, assume_energy_ramp, \
                                      dropna_co2factors)
        name = 'processed_co2_total_{}_{}.parquet'.format(st, et)
        res_str += [name]
        df.to_parquet(os.path.join(cache, name))

    # Load cached results files
    results_df = []
    for name in res_str:
        logger.info(f"Loading results file {name}")
        results_df += [pd.read_parquet(os.path.join(cache, name))]
    
    flatten = pd.concat(results_df, ignore_index=True)
    res = flatten[flatten['Time'].between(start_time, end_time, inclusive="right")]

    logger.info('Completed get_total_emissions_by_DI_DUID')

    if return_all:
        return res
    else:
        if generation_sent_out:
            return res[['DUID', 'Time', 'Region', 'Plant_Emissions_Intensity', 'Energy', 'PCT_AUXILIARY_LOAD', \
                        'Energy_SO', 'Total_Emissions']]
        else:
            return res[['DUID', 'Time', 'Region', 'Plant_Emissions_Intensity', 'Energy', 'Total_Emissions']]


def _generate_timeseries_loop(actual_start, actual_end):
    """Generates a dict of start and end times for looping through the `_total_emissions_process` computation.
    """
    stime = dt.strptime(actual_start, "%Y/%m/%d %H:%M")
    etime = dt.strptime(actual_end, "%Y/%m/%d %H:%M")
    start_series = pd.date_range(stime, etime, freq='MS', normalize=True, inclusive="neither")
    end_series = pd.date_range(stime, etime, freq='MS', normalize=True, inclusive="neither")
    time_segments = {
        'start': [actual_start[0:10]+" 00:00"] + start_series.strftime("%Y/%m/%d %H:%M").to_list(),
        'end': end_series.strftime("%Y/%m/%d %H:%M").to_list() + [actual_end],
        's_str': [stime.strftime("%Y-%m-%d")] + start_series.strftime("%Y-%m-%d").to_list(),
        'e_str': end_series.strftime("%Y-%m-%d").to_list() + [etime.strftime("%Y-%m-%d")]
    }
    return time_segments


def _total_emissions_process(start_time, end_time, cache, filter_regions=None,
                             generation_sent_out=True, assume_energy_ramp=True, dropna_co2factors=True):
    """Process for calculating total emissions based on the parameters defined in `get_total_emissions_by_DI_DUID`.
    """
    # Download Unit Dispatch Data and Generation Information
    disp_df = download_unit_dispatch(start_time, end_time, cache, source_initialmw=False, source_scada=True,
                                     return_all=False, check=False, overwrite="scada", rm_negative=True)

    geninfo_df = download_dudetailsummary(cache)

    # Merge geninfo and filter out loads
    disp_df.set_index('DUID', inplace=True)
    geninfo_df.set_index('DUID', inplace=True)
    filt_df = disp_df.join(geninfo_df[['REGIONID', 'DISPATCHTYPE']], how="left")
    filt_df.reset_index(inplace=True)
    
    filt_df = filt_df[filt_df['DISPATCHTYPE'] == 'GENERATOR']

    # Filter by region if specified
    if filter_regions:
        if not pd.Series(filt_df["REGIONID"].unique()).isin(filter_regions).any():
            raise ValueError("filter_region paramaters passed were not found in NEM regions")
        filt_df = filt_df[filt_df["REGIONID"].isin(filter_regions)]

    # Merge Energy data with Plant Emissions Factors
    filt_df['year'], filt_df['month'] = filt_df['Time'].dt.year, filt_df['Time'].dt.month
    co2factors_df = _get_duid_emissions_intensities(start_time, end_time, cache)
    plt_df = pd.merge(left=filt_df,
                      right=co2factors_df[["file_year", "file_month", "DUID", "CO2E_EMISSIONS_FACTOR"]],
                      left_on=["year", "month", "DUID"],
                      right_on=["file_year", "file_month", "DUID"],
                      how="left")

    # Filter out Data with Null CO2_EMISSIONS_FACTORS
    if dropna_co2factors:
        plt_df = plt_df[~plt_df['CO2E_EMISSIONS_FACTOR'].isna()]

    # Calculate Energy (MWh)
    if not assume_energy_ramp:
        plt_df["Energy"] = plt_df["Dispatch"] * (DISP_INT_LENGTH / 60)
        result = plt_df
    else:
        result = _calculate_energy_ramp(plt_df)
        
    # Calculate Sent-Out Energy (MWh)
    if generation_sent_out:
        result = _calculate_sent_out(result)
        # Compute emissions
        result["Total_Emissions"] = result["Energy_SO"] * result["CO2E_EMISSIONS_FACTOR"]
    else:
        result["Total_Emissions"] = result["Energy"] * result["CO2E_EMISSIONS_FACTOR"]

    result.rename(columns={"CO2E_EMISSIONS_FACTOR": "Plant_Emissions_Intensity",
                           "REGIONID": "Region"}, inplace=True)
    return result


def _get_duid_emissions_intensities(cache, start_time, end_time):
    """Merges emissions factors from GENSETID to DUID and cleans data"""
    co2factors_df = download_plant_emissions_factors(cache, start_time, end_time)
    genset_map = download_genset_map(cache)
    co2factors_df = co2factors_df.merge(right=genset_map[["GENSETID", "DUID"]],
                                        on=["GENSETID"],
                                        how="left")

    # Filter out older assumptions, where duplicate CO2 factors exist for data entry
    co2factors_df['CO2E_DATA_YEAR'] = co2factors_df['CO2E_DATA_SOURCE'].map(CO2E_DATA_SOURCE_YEARMAP)
    co2factors_df = co2factors_df.sort_values(['CO2E_DATA_YEAR'], ascending=True)
    co2factors_df = co2factors_df.drop_duplicates(['file_year', 'file_month', 'CO2E_ENERGY_SOURCE', 'DUID'], keep='last')

    # Find and correct duplicate entries for DUIDs
    co2factors_df = _condense_genset_co2_differences(co2factors_df)

    # Ammend Hydro NaN values to zero
    co2factors_df.loc[(co2factors_df['CO2E_EMISSIONS_FACTOR'].isna()) & \
                    (co2factors_df['CO2E_ENERGY_SOURCE']=='Hydro'), 'CO2E_EMISSIONS_FACTOR'] = 0.0

    return co2factors_df[['file_year', 'file_month', 'DUID', 'CO2E_EMISSIONS_FACTOR', \
                          'CO2E_ENERGY_SOURCE', 'CO2E_DATA_SOURCE']]


def _condense_genset_co2_differences(all_df):
    """Patch duplicate or differing co2 factors for the same year-month-DUID."""
    for duid in all_df[all_df.duplicated(['file_year','file_month','DUID'])]['DUID'].unique():
        correction = all_df[all_df['DUID']==duid].dropna(subset=['CO2E_EMISSIONS_FACTOR', 'CO2E_ENERGY_SOURCE', \
                                                                 'CO2E_DATA_SOURCE'])

        # If year-month-gensetid is duplicated in correction df, average vals and drop correction
        if correction.duplicated(subset=['file_year','file_month','DUID']).any():
            result = []
            for yr in correction['file_year'].unique():
                for mn in correction[correction['file_year'] == yr]['file_month'].unique():
                    subset = correction[(correction['file_year'] == yr) & (correction['file_month'] == mn)]
                    descriptor = [' / '.join(subset['CO2E_ENERGY_SOURCE'].to_list()) \
                        if len(subset['CO2E_DATA_SOURCE'].unique())==1 else subset['CO2E_ENERGY_SOURCE'].iloc[0]][0]
                    emissions = subset['CO2E_EMISSIONS_FACTOR'].mean()
                    subset = subset.drop_duplicates(['file_year','file_month','DUID'], keep='first')
                    subset['CO2E_ENERGY_SOURCE'] = descriptor
                    subset['CO2E_EMISSIONS_FACTOR'] = emissions
                    result += [subset]
            result = pd.concat(result,ignore_index=True)
        else:
            result = correction
        
        # Remove duid entry in main df
        all_df = all_df[~all_df['DUID'].isin([duid])]

        # Add resultant duid entry to main df
        all_df = pd.concat([all_df, result], ignore_index=True)
    return all_df.drop(['GENSETID'], axis=1).reset_index(drop=True)


def _calculate_energy_ramp(dispatch_df):
    """Returns dataframe with energy calculated as ramp between dispatch scada points.
    """
    logger.info('Compiling Energy from Dispatch')
    aggregate = []
    for duid in dispatch_df['DUID'].unique():
        sub_df = dispatch_df[dispatch_df['DUID'] == duid]
        sub_df = sub_df.sort_values('Time')
        sub_df.reset_index(drop=True, inplace=True)
        sub_df['Dispatch_prev'] = np.nan
        sub_df.loc[1:, 'Dispatch_prev'] = sub_df['Dispatch'][:len(sub_df)-1].to_list()
        sub_df.loc[:, 'Energy'] = (0.5*(sub_df['Dispatch'] - sub_df['Dispatch_prev']) + sub_df['Dispatch_prev']) \
            * (DISP_INT_LENGTH / 60)
        aggregate += [sub_df]
    aggregate = pd.concat(aggregate, ignore_index=True)
    return aggregate.reset_index(drop=True)


def _calculate_sent_out(energy_df):
    """Returns dataframe with sent-out generation calculated by considering auxload factor for corresponding DUID.
    """
    logger.info('Compiling Sent Out Generation')
    auxload = read_plant_auxload_csv()

    # Merge data and compute auxilary load factor
    so_df = energy_df.merge(auxload, on=["DUID"], how="left")
    so_df['pct_sent_out'] = (100 - so_df["PCT_AUXILIARY_LOAD"]) / 100
    """add error checking measure for % of no matches in auxload"""
    so_df['pct_sent_out'].fillna(1.0, inplace=True)
    so_df["Energy_SO"] = so_df["Energy"] * so_df['pct_sent_out']
    return so_df


def get_marginal_emitter(cache, start_time, end_time):
    """Retrieves the marginal emissions intensity for each dispatch interval and region. This factor being the weighted
    sum of the generators contributing to price-setting. Although not necessarily common, there may be times where
    multiple technology types contribute to the marginal emissions - note however that the 'DUID' and 'CO2E_ENERGY_SOURCE'
    returned will reflect only the plant which makes the greatest contribution towards price-setting.

    Parameters
    ----------
    cache : str
        Raw data location in local directory
    start_time : str
        Start Time Period in format 'yyyy/mm/dd HH:MM'
    end_time : str
        End Time Period in format 'yyyy/mm/dd HH:MM'

    Returns
    -------
    pandas.DataFrame
        Data is returned as:

        ==================  ========  ==================================================================================================
        Columns:            Type:     Description:
        Time                datetime  Timestamp reported as end of dispatch interval.
        Region              str       The NEM region corresponding to the marginal emitter data. 
        Intensity_Index     float     The intensity index [tCO2e/MWh] (as by weighted contributions) of the price-setting generators.
        DUID                str       Unit identifier of the generator with the largest contribution on the margin for that Time-Region.
        CO2E_ENERGY_SOURCE  str       Unit energy source with the largest contribution on the margin for that Time-Region.
        ==================  ========  ==================================================================================================

    """
    # Check if cache is an existing directory
    hp._check_cache(cache)

    # Download CDEII, Price Setter Files and Generation Information
    ## gen_info = download_generators_info(cache)
    logger.warning('Warning: Gen_info table only has most recent NEM registration and exemption list. Does not account for retired generators')
    co2_factors = _get_duid_emissions_intensities(cache, start_time, end_time)
    price_setters = download_pricesetter_files(cache, start_time, end_time)

    # Drop Basslink
    filt_df = price_setters[~price_setters['Unit'].str.contains('T-V-MNSP1')]
    filt_df = filt_df.rename(columns={'Unit': 'DUID', 'PeriodID': 'Time', 'RegionID': 'Region'})
    filt_df['file_year'], filt_df['file_month'] = filt_df['Time'].dt.year, filt_df['Time'].dt.month

    # Merge CO2 Factors
    filt_df = pd.merge(filt_df[["Time", "Region", "DUID", "Increase", "file_year", "file_month"]],
        co2_factors[["DUID", "file_year", "file_month", "CO2E_EMISSIONS_FACTOR", "CO2E_ENERGY_SOURCE"]],
        how="left", on=["DUID", "file_year", "file_month"], sort=False)

    filt_df.drop(['file_year', 'file_month'], axis=1, inplace=True)

    # Weigh CO2 intensity by 'Increase' contributions
    filt_df['weighted_co2_factor'] = filt_df['Increase'] * filt_df['CO2E_EMISSIONS_FACTOR']

    # Aggregate sum of weighted CO2 intensities
    values = filt_df.groupby(by=['Time','Region'], axis=0).sum()
    values = values.reset_index()[['Time','Region','weighted_co2_factor']]
    values.rename(columns={'weighted_co2_factor': 'Intensity_Index'}, inplace=True)

    # Identify Emissions tech/DUID with the largest contribution (increase) value for Time-Region
    source = filt_df.sort_values(['Increase']).drop_duplicates(['Time','Region'], keep="last")[['Time', 'Region', 'DUID', 'CO2E_ENERGY_SOURCE']]
    result = values.merge(source, on=['Time','Region'], how='left')

    return result


def tech_rename(fuel, tech_descriptor, dispatch_type):
    # """LEGACY. DEPRECATED.
    # Name technology type based on fuel, and descriptions of AEMO

    # Parameters
    # ----------
    # fuel : str
    #     Fuel Source - Descriptor field of a specific DUID
    # tech_descriptor : str
    #     Technology Type - Descriptor field of a specific DUID
    # dispatch_type : str
    #     Dispatch Type field of a specific DUID

    # Returns
    # -------
    # str
    #     Returns a simplified name for the above arguments
    # """

    # name = fuel
    # if fuel in ['Solar', 'Wind', 'Black Coal', 'Brown Coal']:
    #     pass
    # elif tech_descriptor == 'Battery':
    #     if dispatch_type == 'Load':
    #         name = 'Battery Charge'
    #     else:
    #         name = 'Battery Discharge'
    # elif tech_descriptor in ['Hydro - Gravity', 'Run of River']:
    #     name = tech_descriptor
    # elif tech_descriptor == 'Pump Storage':
    #     if dispatch_type == 'Load':
    #         name = 'Pump Storage Charge'
    #     else:
    #         name = 'Pump Storage Discharge'
    # elif tech_descriptor == '-' and fuel == '-' and dispatch_type == 'Load':
    #     name = 'Pump Storage Charge'
    # elif tech_descriptor == 'Open Cycle Gas turbines (OCGT)':
    #     name = 'OCGT'
    # elif tech_descriptor == 'Combined Cycle Gas Turbine (CCGT)':
    #     name = 'CCGT'
    # elif fuel in ['Natural Gas / Fuel Oil', 'Natural Gas'] and tech_descriptor == 'Steam Sub-Critical':
    #     name = 'Gas Thermal'
    # elif isinstance(tech_descriptor, str) and 'Engine' in tech_descriptor:
    #     name = 'Reciprocating Engine'
    # return name
    raise Exception("DEPRECATED in this version of NEMED. Refer to documentation: https://nemed.readthedocs.io/en/latest/")


def aggregate_data_by(data, by):
    """Aggregate the total emissions dataset metrics of Sent Out Generation, Total Emissions and Intensity Index.

    Parameters
    ----------
    data : pandas.DataFrame
        Dataframe input must correspond to the output from `get_total_emissions` with the `by` arugment set to None.
    by : str
        One of ['interval', 'hour', 'day', 'month', 'year']

    Returns
    -------
    pandas.DataFrame
        Data is returned as:

        ===============  ========  ==============================================================================================================================================
        Columns:         Type:     Description:
        TimeBeginning    datetime  Timestamp for start of interval or aggregation period. Only returned if `by` parameter is set.
        TimeEnding       datetime  Timestamp for end of interval or aggregation period.
        Region           str       The NEM region corresponding to data. 'NEM' field reflects all regions and is returned if `filter_regions` is None from `get_total_emissions`. 
        Energy           float     The total (sent-out if `generation_sent_out` is True from `get_total_emissions`) energy for the corresponding region and time.
        Total_Emissions  float     The total emissions for the corresponding region and time.
        Intensity_Index  float     The intensity index as above, considering the total emissions divided by (sent-out) energy.
        ===============  ========  ==============================================================================================================================================

    Raises
    ------
    Exception
        Invalid dataframe input.
    """
    aggregate = []
    for region in data['Region'].unique():
        sub_df = data[data['Region']==region]
        if ('TimeEnding' in sub_df.columns):
            sub_df = sub_df.rename(columns={'TimeEnding': 'Time'})
        
        if ('TimeBeginning' in sub_df.columns):
            raise Exception("already aggregated data cannot be passed to `aggregate_data_by` function. " +\
                "The `get_total_emissions` `by` input must be set to None to use this function post-operand")       

        sub_df = _time_aggregations(data=sub_df, by=by)
        sub_df.insert(2, 'Region', region)
        aggregate += [sub_df]
    aggregate = pd.concat(aggregate, ignore_index=True)
    return aggregate


def _time_aggregations(data, by):
    """Aggregations of total emissions dataset from 5-minute DI resolution to hour, day, month, or year
    """
    result = data.copy()
    en_colname = result.columns[result.columns.str.contains('Energy')][0]
    agg_map = {en_colname: np.sum, "Total_Emissions": np.sum}
    if 'Intensity_Index' in result.columns:
        reproduce_II = True
        result.drop(['Intensity_Index'], axis=1, inplace=True)
    else:
        reproduce_II = False

    # Shift data to time-beginning for aggregations
    result['Time'] = result['Time'] - timedelta(minutes=5)
    result.set_index(['Time'], inplace=True)

    # Aggregations
    if by == "interval":
        # Format result to show on interval resolution
        result.drop(["Region"],axis=1,inplace=True)
        result.insert(0, "TimeBeginning", result.index)
        result.insert(1, "TimeEnding", result.index + timedelta(minutes=5))
        result.reset_index(drop=True, inplace=True)
        
    elif by == "hour":
        # Hourly time resolution
        result = result.groupby(by=[result.index.year,
                                    result.index.month,
                                    result.index.day,
                                    result.index.hour]).agg(agg_map)

        result.index.names = ['Y','m','d','H']
        result.reset_index(inplace=True)
        result.insert(0,"TimeBeginning", pd.to_datetime(dict(year=result.Y,
                                                             month=result.m,
                                                             day=result.d,
                                                             hour=result.H)))
        result.drop(['Y','m','d','H'], axis=1, inplace=True)
        result.insert(1,"TimeEnding", result['TimeBeginning'] + timedelta(hours=1))

    elif by == "day":
        # Daily time resolution
        result = result.groupby(by=[result.index.year,
                                    result.index.month,
                                    result.index.day]).agg(agg_map)

        result.index.names = ['Y','m','d']
        result.reset_index(inplace=True)
        result.insert(0,"TimeBeginning", pd.to_datetime(dict(year=result.Y,
                                                             month=result.m,
                                                             day=result.d)))
        result.drop(['Y','m','d'], axis=1, inplace=True)
        result.insert(1, "TimeEnding", result['TimeBeginning'] + timedelta(days=1))

    elif by == "month":
        # Monthly time resolution
        result = result.groupby(by=[result.index.year,
                                    result.index.month]).agg(agg_map)
        result.index.names = ['Y','m']
        result.reset_index(inplace=True)
        result.insert(0,"TimeBeginning", pd.to_datetime(dict(year=result.Y,
                                                             month=result.m,
                                                             day=1)))
        result.insert(1,"TimeEnding", pd.to_datetime(dict(year=result.Y,
                                                          month=result.m + 1,
                                                          day=result['TimeBeginning'].dt.day)))
        result.drop(['Y','m'], axis=1, inplace=True)

    elif by == "year":
        # Annual time resolution
        result = result.groupby(by=[result.index.year]).agg(agg_map)
        result.index.names = ['Y']
        result.reset_index(inplace=True)
        result.insert(0,"TimeBeginning", pd.to_datetime(dict(year=result.Y,
                                                             month=1,
                                                             day=1)))
        result.insert(1,"TimeEnding", pd.to_datetime(dict(year=result.Y + 1,
                                                          month=1,
                                                          day=1)))
        result.drop(['Y'], axis=1, inplace=True)

    else:
        raise Exception(
            "Error: invalid by argument. Must be one of [interval, hour, day, month, year]"
        )

    # Update Intensity Index is found in data input
    if reproduce_II:
        result['Intensity_Index'] = result['Total_Emissions'] / result[en_colname]
        result['Intensity_Index'] = result['Intensity_Index'].fillna(0.0)

    return result.round(3)


""" Modifies functionality from nempy.historical_inputs.xml_cache"""
from pathlib import Path
import requests
import zipfile
import io
import os
import glob
import json
import xmltodict
import logging
import pandas as pd
from tqdm import tqdm
from datetime import datetime, timedelta
from nempy.historical_inputs.xml_cache import XMLCacheManager
logger = logging.getLogger(__name__)


def modpricesetter_get_file_name(self):
    """Modified function from nempy.historical_inputs.xml_cache
    """
    year, month, day = self._get_market_year_month_day_as_str()
    interval_number = self._get_interval_number_as_str()
    base_name = "NEMPriceSetter_{year}{month}{day}{interval_number}00.xml"
    name = base_name.format(
        year=year, month=month, day=day, interval_number=interval_number
    )
    path_name = Path(self.cache_folder) / name
    name_OCD = name.replace(".xml", "_OCD.xml")
    path_name_OCD = Path(self.cache_folder) / name_OCD
    if os.path.exists(path_name):
        return name
    elif path_name_OCD:
        return name_OCD
    else:
        return name


def modpricesetter_download_xml_from_nemweb(self):
    """Modified function from nempy.historical_inputs.xml_cache
    """
    year, month, day = self._get_market_year_month_day_as_str()
    base_url = "https://www.nemweb.com.au/Data_Archive/Wholesale_Electricity/NEMDE/{year}/NEMDE_{year}_{month}/" + \
        "NEMDE_Market_Data/NEMDE_Files/NemPriceSetter_{year}{month}{day}_xml.zip"
    url = base_url.format(year=year, month=month, day=day)
    r = requests.get(url)
    z = zipfile.ZipFile(io.BytesIO(r.content))
    z.extractall(self.cache_folder)


def overwrite_xmlcachemanager_with_pricesetter_config():
    """Overwrites nempy xml_cache with modified functions to pull price setter data from AEMO NEMWEB.
    """
    # XMLCacheManager.get_file_name = modpricesetter_get_file_name(xmlcachemanager)
    # XMLCacheManager._download_xml_from_nemweb = modpricesetter_download_xml_from_nemweb(xmlcachemanager)
    XMLCacheManager.get_file_name = modpricesetter_get_file_name
    XMLCacheManager._download_xml_from_nemweb = modpricesetter_download_xml_from_nemweb

def convert_xml_to_json(cache, start_date_str, end_date_str, clean_up=False):
    """Converts all XML files found in cache to JSON format. Best to use an entirely separate cache folder from other
    nemosis or nempy projects!

    Parameters
    ----------
    cache : str
        Defined folder in directory to use as cache.
    clean_up : bool, optional
        Setting to True will remove XML files once they have been converted to JSON, by default False.
    """
    # Establish daterange
    sdate = datetime.strptime(start_date_str, "%Y/%m/%d")
    edate = datetime.strptime(end_date_str, "%Y/%m/%d")
    date_str_list = [datetime.strftime(i,"%Y%m%d") for i in pd.date_range(sdate,edate)]

    # Find only a subset of XML files in the given daterange
    xml_subset_nested = [glob.glob(os.path.join(cache, "NEMPriceSetter_{}*.xml".format(i))) for i in date_str_list]
    xml_subset = [item for sublist in xml_subset_nested for item in sublist]
    xml_files = glob.glob(os.path.join(cache, "NEMPriceSetter_*.xml"))
    print("Converting selected {} XML files to JSON, of {} cached files".format(len(xml_subset),len(xml_files)))
    print(xml_subset)
    # Read XML files and convert to JSON
    for filename in tqdm(xml_subset):
        handle = open(filename, 'r')
        content = handle.read()
        d = xmltodict.parse(content)
        write_file = filename.replace(".xml", ".json")
        json_path = write_file
        with open(json_path, 'w') as fp:
            json.dump(d['SolutionAnalysis']['PriceSetting'], fp)

    # Remove XML files if clean_up
    if clean_up:
        print("Clearing {} XML files from cache".format(len(xml_files)))
        for filename in xml_files:
            os.remove(os.path.join(cache, filename))


def read_json_to_df(cache, start_dt, end_dt):
    """Reads JSON files found in cache and returns price setter data as pandas dataframe.

    Parameters
    ----------
    cache : str
        Defined folder in directory to use as cache.

    Returns
    -------
    pd.DataFrame
        Price Setter dataframe containing columns: [PeriodID, RegionID, Market, Price, DUID, DispatchedMarket, BandNo,
        Increase, RRNBandPrice, BandCost]
    """
    # Establish files daterange
    collect_sdt = start_dt
    if (end_dt.hour == 0) & (end_dt.minute == 0):
        collect_edt = end_dt
    else:
        collect_edt = end_dt + timedelta(days=1)

    date_str_list = [datetime.strftime(i,"%Y-%m-%d") for i in pd.date_range(collect_sdt, collect_edt)]

    # Find only a subset of JSON files in the given daterange
    JSON_subset = [glob.glob(os.path.join(cache, "NEMED_PS_DAILY_{}*.json".format(i))) for i in date_str_list]
    JSON_subset = [item for sublist in JSON_subset for item in sublist]

    print("Reading selected {} JSON files to pandas, of cached files".format(len(JSON_subset)))
    logger.info("Loading Cached Price Setter Files...")

    all_df = []
    for file in tqdm(JSON_subset):
        with open(file, 'r') as f:
            data = json.loads(f.read())
        df_nested_list = pd.json_normalize(data)
        df_nested_list['@PeriodID'] = pd.to_datetime(df_nested_list['@PeriodID'], format="%Y-%m-%d %H:%M:%S").dt.tz_localize(None)
        df_nested_list = df_nested_list[(df_nested_list['@Market'] == 'Energy') &
                                        (df_nested_list['@DispatchedMarket'] == 'ENOF')]
        all_df += [df_nested_list]

    all_df = pd.concat(all_df)
    all_df.columns = all_df.columns.str.strip('@')
    all_df.drop(['Market','DispatchedMarket'], axis=1, inplace=True)
    all_df = all_df.astype({'RegionID': str, 'Price': float, 'Unit': str, 'BandNo': int, \
                            'Increase': float, 'RRNBandPrice': float, 'BandCost': float})
    all_df = all_df[all_df['PeriodID'].between(start_dt, end_dt, inclusive="right")].sort_values(['PeriodID','RegionID'])
    return all_df.reset_index(drop=True)



""" Downloader functions for retrieving data from various sources"""
from nemosis import dynamic_data_compiler, static_table
from nemosis.data_fetch_methods import _read_mms_csv, _dynamic_data_fetch_loop
from nempy.historical_inputs.xml_cache import XMLCacheManager as XML
from .defaults import *
from .helper_functions import helpers as hp
from .helper_functions.mod_xml_cache import overwrite_xmlcachemanager_with_pricesetter_config, convert_xml_to_json,\
    read_json_to_df, modpricesetter_get_file_name, modpricesetter_download_xml_from_nemweb
from .helper_functions.mod_nemosis import overwrite_nemosis_defaults, mod_dynamic_data_fetch_loop

import logging
import os
import json
import glob
from tqdm import tqdm
import numpy as np
import pandas as pd
from datetime import datetime, timedelta
import requests
from pathlib import Path
DISPATCH_INT_MIN = 5
logger = logging.getLogger(__name__)


def download_cdeii_table():
    # """LEGACY. DEPRECATED.
    
    # Retrieves the most recent Carbon Emissions Factor data per generation unit (DUID) published to CDEII dataset in
    # AEMO NEMWEB.

    # .. warning::
    #     This CDEII table ('Available Generators File') is only the most recent generating units from AEMO's current
    #     CDEII reporting week. Attempting to retrieve older historical data may lead to missing generating unit emissions
    #     factors.

    # Returns
    # -------
    # pandas.DataFrame
    #     AEMO CDEII data containing columns=["STATIONNAME","DUID","REGIONID","CO2E_EMISSIONS_FACTOR",
    #     "CO2E_ENERGY_SOURCE","CO2E_DATA_SOURCE"]. CO2E_EMISSIONS_FACTOR is a measure in tCO2-e/MWh
    # """
    # table = _read_mms_csv(CDEII_URL, usecols=[4, 5, 7, 8, 9, 10])
    # return table
    raise Exception("DEPRECATED in this version of NEMED.")



def download_generators_info(cache):
    """Retrieves the Generators and Scheduled Loads static table via NEMOSIS (published by AEMO in NEM Registration and
    Exemption List file). Data reflects the most recent file uploaded by AEMO.

    .. warning::
        This Generators and Scheduled Load table is only the most recent data and is a static file

    Parameters
    ----------
    cache : str
        Raw data location in local directory.

    Returns
    -------
    pandas.DataFrame
        AEMO data containing columns=['Participant', 'Station Name', 'Region', 'Dispatch Type', 'Category',
       'Classification', 'Fuel Source - Primary', 'Fuel Source - Descriptor',
       'Technology Type - Primary', 'Technology Type - Descriptor',
       'Aggregation', 'DUID', 'Reg Cap (MW)']
    """
    cache = hp._check_cache(cache)
    table = static_table(table_name="Generators and Scheduled Loads", raw_data_location=cache)
    return table


def download_duid_auxload():
    # """DEPRECATED: Replaced by read_plant_auxload_csv()
    # Retrieves auxilary load data from static AEMO's ISP assumptions.

    # .. info::
    #     This data is obtained from a static datasource for which the majority of plant data is obtained from the 2020
    #     ISP. More recent data published by AEMO does not reflect unit-level, rather only fuel/technology type
    #     assumptions. Refer to NEMED documentation for more details.

    # Returns
    # -------
    # pandas.DataFrame
    #     AEMO data containing columns=['EFFECTIVEFROM', 'DATASOURCE', 'DUID', 'GENERATOR', 'PCT_AUXILIARY_LOAD', 'NOTE']
    # """
    # map = _download_duid_mapping()
    # auxload = _download_iasr_existing_gens()
    # merged_table = pd.merge(map, auxload, on='Generator', how='left')
    # return merged_table
    raise Exception("DEPRECATED in this version of NEMED. See `read_plant_auxload_csv`")


def read_plant_auxload_csv(select_columns=['EFFECTIVEFROM', 'DUID', 'PCT_AUXILIARY_LOAD'],
                            coltype={'EFFECTIVEFROM': str, 'DUID': str, 'PCT_AUXILIARY_LOAD': float}):
    """Reads locally stored .csv in package with auxiliary load data mapped to each DUID. Users can update this .csv
    with custom/missing values should they wish.

    Parameters
    ----------
    select_columns : list, optional
        Columns of the dataset to return, by default ['EFFECTIVEFROM', 'DUID', 'PCT_AUXILIARY_LOAD']
    coltype : dict, optional
        Datatype corresponding to each field, by default {'EFFECTIVEFROM': datetime, 'DUID': str,
        'PCT_AUXILIARY_LOAD': float}

    Returns
    -------
    pandas.DataFrame
        Custom table containing columns=['EFFECTIVEFROM', 'DUID', 'PCT_AUXILIARY_LOAD']
    """
    filepath = Path(__file__).parent / "./data/plant_auxiliary/_plant_auxload_assumptions.csv"
    table = pd.read_csv(filepath)
    table['EFFECTIVEFROM'] = pd.to_datetime(table['EFFECTIVEFROM'], format="%d/%m/%Y")
    return table[table.columns[table.columns.isin(select_columns)]]


def download_plant_emissions_factors(cache, start_date, end_date):
    """Retrieves CO2-equivalent emissions intensity factors (tCO2-e/MWh) for each generator. Metric is reflective of
    sent-out generation. Underlying data is sourced from the 'GENUNITS' table of AEMO MMS at monthly time resolution.

    Parameters
    ----------
    cache : str
        Raw data location in local directory
    start_date : str
        Data download period start, in the format: 'yyyy/mm/dd HH:MM'
    end_date : str
        Data download period end, in the format: 'yyyy/mm/dd HH:MM'

    Returns
    -------
    pandas.DataFrame
        Plant Emissions Factor Data with columns=['file_year', 'file_month', 'GENSETID', 'CO2E_EMISSIONS_FACTOR',
        'CO2E_ENERGY_SOURCE', 'CO2E_DATA_SOURCE'] 

    Raises
    ------
    Exception
        Data Unavailable for dates prior 05-2011
    """
    # Input Validation
    hp._validate_variable_type(start_date, str, "start_date")
    hp._validate_variable_type(end_date, str, "end_date")
    start_date = hp._validate_and_convert_date(start_date, "start_date")
    end_date = hp._validate_and_convert_date(end_date, "end_date")
    cache = hp._check_cache(cache)

    overwrite_nemosis_defaults()

    if start_date < datetime(2011,5,1):
        raise Exception("DATA UNAVAILABLE: GENUNITS table is not found in MMS prior to 05-2011. \
                        Unit emissions factors cannot be obtained.")

    # Data Retrieval
    df = mod_dynamic_data_fetch_loop(start_search=start_date,
                                    start_time=start_date,
                                    end_time=end_date,
                                    table_name="GENUNITS",
                                    raw_data_location=cache,
                                    select_columns=["GENSETID", "CO2E_EMISSIONS_FACTOR", "CO2E_ENERGY_SOURCE", \
                                        'CO2E_DATA_SOURCE'],
                                    date_filter=None,
                                    fformat="feather",
                                    )
    df = pd.concat(df)
    df['file_year'] = df['file_year'].astype(int)
    df['file_month'] = df['file_month'].astype(int)
    df['CO2E_EMISSIONS_FACTOR'] = df['CO2E_EMISSIONS_FACTOR'].astype(float)
    return df.sort_values(['GENSETID','file_year','file_month'])


def download_genset_map(cache, asof_date=None):
    """Download the GENSETID to DUID mapping from DUALLOC MMS Table.

    Parameters
    ----------
    cache : str
        Raw data location in local directory
    asof_date : str, optional
        Date to retrieve DUALLOC table as of, in the format: 'yyyy/mm/dd HH:MM', by default None which will retrieve
        recent data 

    Returns
    -------
    pandas.DataFrame

        =============  =====  ================================================
        Columns:       Type:  Description:
        EFFECTIVEDATE  str    Effective Date as defined by AEMO.
        DUID           str    Dispatchable Unit Identifier as defined by AEMO.
        GENSETID       str    Generator Set Identifier as defined by AEMO.
        =============  =====  ================================================

    Raises
    ------
    Exception
        Parameter `asof_date` exceeds the earliest available DUALLOC table in MMS
    """
    if asof_date != None:
        latest = hp._validate_and_convert_date(asof_date, "asof_date")
    else:
        latest = datetime(datetime.now().year, datetime.now().month, 1) - timedelta(days = 90)
    
    cache = hp._check_cache(cache)
    overwrite_nemosis_defaults()

    if latest < datetime(2020,10,1):
        raise Exception("DATA UNAVAILABLE: DUALLOC table is not found in MMS prior to 10-2020. " + \
                        "Cannot correctly map emissions factors. Retry function without specifying `asof_date`")

    # Data Retrieval
    df = mod_dynamic_data_fetch_loop(start_search=latest,
                                    start_time=latest - timedelta(hours=1),
                                    end_time=latest,
                                    table_name="DUALLOC",
                                    raw_data_location=cache,
                                    select_columns=["EFFECTIVEDATE", "DUID", "GENSETID", "LASTCHANGED"],
                                    date_filter=None,
                                    fformat="feather",
                                    )
    df = pd.concat(df)
    df = df.drop(['file_year','file_month', 'LASTCHANGED'], axis=1)
    filtered = df.sort_values('EFFECTIVEDATE').drop_duplicates(['GENSETID'], keep='last')
    return filtered.sort_values(['GENSETID','EFFECTIVEDATE']).reset_index(drop=True)


def download_dudetailsummary(cache, asof_date=None):
    """Download the DUDETAILSUMMARY MMS table with mapping of Dispatch Type and Region to DUID

    Parameters
    ----------
    cache : str
        Raw data location in local directory
    asof_date : str, optional
        Date to retrieve DUALLOC table as of, in the format: 'yyyy/mm/dd HH:MM', by default None which will retrieve
        recent data 

    Returns
    -------
    pandas.DataFrame

        ============  =====  ================================================
        Columns:      Type:  Description:
        DUID          str    Dispatchable Unit Identifier as defined by AEMO.
        START_DATE    str    Date of data entry as defined by AEMO.
        DISPATCHTYPE  str    Dispatch Type of DUID as 'GENERATOR' or 'LOAD'.
        REGIONID      str    Region of DUID.
        ============  =====  ================================================

    Raises
    ------
    Exception
        Parameter `asof_date` exceeds the earliest available DUALLOC table in MMS
    """
    if asof_date != None:
        latest = hp._validate_and_convert_date(asof_date, "asof_date")
    else:
        latest = datetime(datetime.now().year, datetime.now().month, 1) - timedelta(days = 90)
    
    cache = hp._check_cache(cache)
    overwrite_nemosis_defaults()

    if latest < datetime(2009,7,1):
        raise Exception("DATA UNAVAILABLE: DUDETAILSUMMARY table is not found in MMS prior to 07-2009. " + \
                        "Cannot correctly map emissions factors. Retry function without specifying `asof_date`")

    # Data Retrieval
    df = mod_dynamic_data_fetch_loop(start_search=latest,
                                    start_time=latest - timedelta(hours=1),
                                    end_time=latest,
                                    table_name="DUDETAILSUMMARY",
                                    raw_data_location=cache,
                                    select_columns=["START_DATE", "DUID", "DISPATCHTYPE", "REGIONID", "LASTCHANGED"],
                                    date_filter=None,
                                    fformat="feather",
                                    )
    df = pd.concat(df)
    df = df.drop(['file_year','file_month', 'LASTCHANGED'], axis=1)
    filtered = df.sort_values('START_DATE').drop_duplicates(['DUID'], keep='last')
    return filtered.sort_values(['DUID','START_DATE']).reset_index(drop=True)


def _download_duid_mapping():
    # """LEGACY. TO BE DEPRECATED.
    
    # A manual record of duid mappings which match unit-level DUIDs to station-level generation names in the IASR
    # dataset.

    # Returns
    # -------
    # pandas.DataFrame
    #     Custom table containing columns=['DUID', 'Generator']
    # """
    # filepath = Path(__file__).parent / "./data/duid_mapping.csv"
    # table = pd.read_csv(filepath)[['DUID', '2021-22-IASR_Generator']]
    # table.columns = ['DUID', 'Generator']
    # return table
    raise Exception("DEPRECATED in this version of NEMED. Use `download_genset_map`")


def _download_iasr_existing_gens(select_columns=['Generator', 'Auxiliary Load (%)'], coltype={'Generator': str,
                                 'Auxiliary Load (%)': float}):
    # """LEGACY. TO BE DEPRECATED.
    
    # Retrieves a static data of the 2021 AEMO IASR Gen Data Summary.

    # Parameters
    # ----------
    # select_columns : list(str), optional
    #     Column names from the IASR to retrieve, by default ['Generator', 'Auxiliary Load (%)']
    # coltype : dict, optional
    #     Data types to use for the selected columns, by default {'Generator': str, 'Auxiliary Load (%)': float}

    # Returns
    # -------
    # pandas.DataFrame
    #     Table extract of IASR containing columns specified as `select_columns`
    # """
    # filepath = Path(__file__).parent / "./data/existing_gen_data_summary.csv"
    # table = pd.read_csv(filepath, dtype=coltype)
    # table = table[table.columns[table.columns.isin(select_columns)]]
    # return table
    raise Exception("DEPRECATED in this verion of NEMED")


def download_aemo_cdeii_summary(filter_start, filter_end, cache):
    """Downloads and combines selected AEMO CDEII Summary Files for a specified date range. The files processed here are
    available from:
    https://aemo.com.au/en/energy-systems/electricity/national-electricity-market-nem/market-operations/settlements-and-payments/settlements/carbon-dioxide-equivalent-intensity-index

    Parameters
    ----------
    filter_start : str
        Data download period start, in the format: 'yyyy/mm/dd HH:MM:SS'
    filter_end : str
        Data download period end, in the format: 'yyyy/mm/dd HH:MM:SS'
    cache : str
        Raw data location in local directory

    Returns
    -------
    pd.DataFrame
        AEMO data containing columns = 'SETTLEMENTDATE', 'REGIONID', 'TOTAL_SENT_OUT_ENERGY',
       'TOTAL_EMISSIONS', 'CO2E_INTENSITY_INDEX']

    Raises
    ------
    ValueError
        Where 'filter_start' exceeds the earliest available CDEII data file supported by NEMED.
    """
    cache = hp._check_cache(cache)

    # Check filter start and end in defaults files datarange.
    fil_start_dt = datetime.strptime(filter_start, "%Y/%m/%d %H:%M")
    fil_end_dt = datetime.strptime(filter_end, "%Y/%m/%d %H:%M")

    default_start_yearname, default_end_yearname = list(CDEII_SUMFILES)[0], list(CDEII_SUMFILES)[-1]

    if fil_start_dt < datetime.strptime(CDEII_SUMFILES[default_start_yearname]['start'],
                                        CDEII_SUMFILES_DTFMT[default_start_yearname]):
        raise ValueError("'filter_start' date exceeds the available CDEII data backdated to {}"\
                            .format(CDEII_SUMFILES[default_start_yearname]['start']))
    elif (fil_end_dt > datetime.strptime(CDEII_SUMFILES[default_end_yearname]['end'],
                                        CDEII_SUMFILES_DTFMT[default_end_yearname])):
        extract_current = True
        if fil_start_dt >= datetime.strptime(CDEII_SUMFILES[default_end_yearname]['start'],
                                        CDEII_SUMFILES_DTFMT[default_end_yearname]):
            extract_historical = False
        else:
            extract_historical = True
    else:
        extract_historical = True
        extract_current = False

    aemodata = []
    if extract_historical:
        # Extract CDEII datafiles for historical
        files_sdt = [datetime.strptime(CDEII_SUMFILES[i]['start'], CDEII_SUMFILES_DTFMT[i]) \
                     for i in CDEII_SUMFILES.keys()]
        files_edt = [datetime.strptime(CDEII_SUMFILES[i]['end'], CDEII_SUMFILES_DTFMT[i]) \
                     for i in CDEII_SUMFILES.keys()]

        # Find nearest dataseries start date
        if np.searchsorted(files_sdt, fil_start_dt)==0:
            extract_from_idx = np.searchsorted(files_sdt, fil_start_dt)
        else :
            extract_from_idx = (np.searchsorted(files_sdt, fil_start_dt)-1)
        ## extract_from = list(CDEII_SUMFILES)[extract_from_idx]

        # Find nearest dataseries end date
        if np.searchsorted(files_edt, fil_end_dt)==0:
            extract_to_idx = np.searchsorted(files_edt, fil_end_dt)+1
        else :
            extract_to_idx = (np.searchsorted(files_edt, fil_end_dt))

        # If date falls in current datafile (not historical)
        if extract_to_idx > len(list(CDEII_SUMFILES)) - 1:
            extract_to_idx = (len(list(CDEII_SUMFILES)) - 1)
        ## extract_to = list(CDEII_SUMFILES)[extract_to_idx]

        # Extract Datafile from AEMO
        for idx in range(extract_from_idx, extract_to_idx+1):
            yearname = list(CDEII_SUMFILES)[idx]
            year = CDEII_SUMFILES[yearname]['year']
            urlbase = CDEII_SUMFILES[yearname]['url']
            # Manual request for change in file naming on aemo website
            if year == '2015':
                url = urlbase
            elif urlbase == 'http://nemweb.com.au/Reports/Current/CDEII/':
                url = urlbase + f"CO2EII_SUMMARY_RESULTS_{year}.CSV"
            else:
                url = urlbase + f'{year}/co2eii_summary_results_{yearname}.csv?la=en'

            filepath = os.path.join(cache, f'AEMO_CO2EII_{yearname}.csv')
            if os.path.exists(filepath):
                os.remove(filepath)
            r = requests.get(url, headers=REQ_URL_HEADERS)
            with open(filepath, 'wb') as f:
                f.write(r.content)
            aemo_file = pd.read_csv(filepath, header=1, usecols=[6, 7, 8, 9, 10])
            aemo_file['SETTLEMENTDATE'] = pd.to_datetime(aemo_file['SETTLEMENTDATE'],
                                                         format=CDEII_SUMFILES_DTFMT[yearname])
            aemodata += [aemo_file]

    if extract_current:
        # Extract CDEII datafiles from current file
        print(f"Extracting AEMO CDEII Datafile for: CURRENT")
        url = "https://www.nemweb.com.au/Reports/Current/CDEII/CO2EII_SUMMARY_RESULTS.CSV"
        filepath = os.path.join(cache, f'AEMO_CO2EII_CURRENT.csv')
        r = requests.get(url, headers=REQ_URL_HEADERS)
        with open(filepath, 'wb') as f:
            f.write(r.content)
        aemo_file = pd.read_csv(filepath, header=1, usecols=[6, 7, 8, 9, 10])
        aemo_file['SETTLEMENTDATE'] = pd.to_datetime(aemo_file['SETTLEMENTDATE'], format="%Y/%m/%d %H:%M:%S")
        aemodata += [aemo_file]

    table = pd.concat(aemodata)
    table = table[table['SETTLEMENTDATE'].between(fil_start_dt, fil_end_dt, inclusive="left")]
    if max(table['SETTLEMENTDATE']) < fil_end_dt:
        print("WARNING: 'filter_end' date exceeds available data from CURRENT CDEII. Latest available data is {}"\
            .format(max(table['SETTLEMENTDATE'])))
    return table.sort_values('SETTLEMENTDATE').reset_index(drop=True)


def download_current_aemo_cdeii_summary(filter_start, filter_end, financialyear="1920"):
    # """
    # LEGACY. TO BE DEPRECATED.
    # .. warning::
    #     TO BE DEPRECATED IN FUTURE VERSIONS. Replaced by **download_aemo_cdeii_summary**
   
    # Retrieve the AEMO CDEII daily summary file by financial year.

    # Parameters
    # ----------
    # filter_start : str
    #     Data download period start, in the format: 'yyyy/mm/dd HH:MM:SS'
    # filter_end : str
    #     Data download period end, in the format: 'yyyy/mm/dd HH:MM:SS'
    # financialyear : str, optional
    #     The financial year to get the cdeii file for, one of ['1920','2122'], by default "1920"

    # Returns
    # -------
    # _type_
    #     _description_
    # """
    # assert(financialyear in ['1920', '2122']), "Financial Year must be one of ['1920','2122']"
    # filepath = Path(__file__).parent / f"./data/CO2EII_SUMMARY_RESULTS_FY{financialyear}.csv"

    # aemo = pd.read_csv(filepath, header=1, usecols=[6, 7, 8, 9, 10])
    # aemo['SETTLEMENTDATE'] = pd.to_datetime(aemo['SETTLEMENTDATE'], format="%d/%m/%Y %H:%M")
    # table = aemo[aemo['SETTLEMENTDATE'].between(filter_start, filter_end)]
    # return table
    raise Exception("DEPRECATED in this version of NEMED. Use `download_aemo_cdeii_summary`")


def get_aemo_comparison_data(filter_start, filter_end, filename='AEMO_CO2EII_August_2022_dataset.csv'):
    # # Call the download func.

    # filepath = Path(__file__).parent / f"../../data/{filename}"
    # aemo = pd.read_csv(filepath, header=1, usecols=[6, 7, 8, 9, 10])

    # fil_start_dt = datetime.strptime(filter_start, "%Y/%m/%d %H:%M:%S")
    # fil_end_dt = datetime.strptime(filter_end, "%Y/%m/%d %H:%M:%S")

    # aemo['SETTLEMENTDATE'] = pd.to_datetime(aemo['SETTLEMENTDATE'], format="%d/%m/%Y %H:%M")
    # table = aemo[aemo['SETTLEMENTDATE'].between(fil_start_dt, fil_end_dt)]
    # return table
    raise Exception("DEPRECATED in this version of NEMED.")


def download_unit_dispatch(start_time, end_time, cache, source_initialmw=False, source_scada=True, overwrite='scada',
                           return_all=True, check=True, rm_negative=True):
    """Downloads historical generation dispatch data via NEMOSIS.

    Parameters
    ----------
    start_time : str
        Start Time Period in format 'yyyy/mm/dd HH:MM'
    end_time : str
        End Time Period in format 'yyyy/mm/dd HH:MM'
    cache : str
        Raw data location in local directory
    source_initialmw : bool
        Whether to download initialmw column from DISPATCHLOAD table, by default False
    source_scada : bool
        Whether to download scada column from DISPATCH_UNIT_SCADA table, by default True
    overwrite : str
        The data value to overwrite in the returned 'Dispatch' column if there is a discrepency in initialmw and scada.
        Must be one of ['initialmw','scada','average']. If one of source_initialmw or source_scada is False, `overwrite`
        has null effect. By default 'scada'.
    return_all : bool
        Whether to return all columns or only ['Time','DUID','Dispatch'], by default False.
    check : bool
        Whether to check for, and remove duplicates after function is complete, by default True.
    rm_negative: bool
        Checks for negative dispatch values in SCADA and replaces them with zero, by default True.

    Returns
    -------
    pd.DataFrame
        Returns generation data as per NEMOSIS

    """
    # Check inputs
    cache = hp._check_cache(cache)
    assert(isinstance(start_time, str)), "`start_time` must be a string in format yyyy/mm/dd HH:MM"
    assert(isinstance(end_time, str)), "`end_time` must be a string in format yyyy/mm/dd HH:MM"
    assert(isinstance(overwrite, (str, type(None)))), "`overwrite` must be a string; one of ['initialmw','scada',\
           'average']"
    if overwrite:
        assert(overwrite in ['initialmw', 'scada', 'average']), "`overwrite` must be a string; one of ['initialmw',\
               'scada', 'average']"

    # Adjust timestamps for Scada interval-beginning
    shift_stime = datetime.strptime(start_time, "%Y/%m/%d %H:%M")
    shift_stime = shift_stime + timedelta(minutes=DISPATCH_INT_MIN)
    shift_etime = datetime.strptime(end_time, "%Y/%m/%d %H:%M")
    shift_etime = shift_etime + timedelta(minutes=DISPATCH_INT_MIN)
    get_start_time = datetime.strftime(shift_stime, "%Y/%m/%d %H:%M:%S")
    get_end_time = datetime.strftime(shift_etime, "%Y/%m/%d %H:%M:%S")

    # Download Dispatch Load table via NEMOSIS
    if source_initialmw:
        disp_load = dynamic_data_compiler(
            start_time=get_start_time,
            end_time=get_end_time,
            table_name="DISPATCHLOAD",
            raw_data_location=cache,
            select_columns=["SETTLEMENTDATE", "DUID", "INITIALMW", "INTERVENTION"],
            fformat="feather",
        )
        disp_load["Time"] = disp_load["SETTLEMENTDATE"] - timedelta(minutes=DISPATCH_INT_MIN)

    elif (not source_initialmw) and (not source_scada):
        raise Exception("No source selected. At least one of `source_initialmw` or `source_scada` must be set True")

    # Download Dispatch Unit Scada table via NEMOSIS (this includes Non-Scheduled generators)
    if source_scada:
        disp_scada = dynamic_data_compiler(
            start_time=get_start_time,
            end_time=get_end_time,
            table_name="DISPATCH_UNIT_SCADA",
            raw_data_location=cache,
            select_columns=["SETTLEMENTDATE", "DUID", "SCADAVALUE"],
            fformat="feather",
        )
        disp_scada["Time"] = disp_scada["SETTLEMENTDATE"] - timedelta(minutes=DISPATCH_INT_MIN)

    """ ====> Change this back to fill dispatch col with initialmw and scada and dispatch col, return all """
    # Adjust for value from the beginning of the interval, to match reporting end of interval

    # Merge Dispatch Load and Scada tables
    if source_scada and source_initialmw:
        disp_load2 = _check_interventions(disp_load)
        disp_load2 = _clean_duplicates(disp_load2, value_col="INITIALMW")

        master = pd.merge(left=disp_load2[['Time', 'DUID', 'INITIALMW']],
                          right=disp_scada[['Time', 'DUID', 'SCADAVALUE']],
                          on=['Time', 'DUID'],
                          how='outer')
        master.sort_values(by=['Time', 'DUID'], inplace=True)
        master.reset_index(drop=True, inplace=True)
        # Compare values and fill
        master['Dispatch'] = np.nan
        # Use whichever value is available
        master['Dispatch'] = np.where(master["INITIALMW"].isnull(), master['SCADAVALUE'], master['Dispatch'])
        master['Dispatch'] = np.where(master["SCADAVALUE"].isnull(), master['INITIALMW'], master['Dispatch'])
        # Check both values match, if available
        master['Dispatch'] = np.where(abs(master['INITIALMW'] - master['SCADAVALUE']) < 1, master['SCADAVALUE'],
                                      master['Dispatch'])
    elif source_scada:
        master = disp_scada[['Time', 'DUID', 'SCADAVALUE']]
        master['Dispatch'] = master['SCADAVALUE']
    elif source_initialmw:
        master = disp_load[['Time', 'DUID', 'INTERVENTION', 'INITIALMW']]
        master['Dispatch'] = master['INITIALMW']

    # Report Error Discrepency (if any), overwrite if specified
    if (not overwrite) and (not master[master['Dispatch'].isnull()].empty):
        print("ERROR DISCREPENCY between SCADAVALUE and INITIALMW. No action performed (values likely to be dropped)" +
              " Check data using download_unit_dispatch() and set `return_all`=True, `check`=False")
    elif overwrite and (not master[master['Dispatch'].isnull()].empty):
        print("ERROR DISCREPENCY between SCADAVALUE and INITIALMW. OVERWRITING using {}".format(overwrite))
        master.reset_index(drop=True, inplace=True)
        if overwrite == "initialmw":
            series = master['INITIALMW']
        elif overwrite == "scada":
            series = master['SCADAVALUE']
        elif overwrite == "average":
            series = (master['INITIALMW'] + master['SCADAVALUE']) / 2

        master['Dispatch'] = np.where(abs(master['INITIALMW'] - master['SCADAVALUE']) >= 1, series,
                                      master['Dispatch'])

    # Final check for intervention periods and duplicates entries
    if check:
        # final = _check_interventions(master)
        final = _clean_duplicates(master)
    else:
        final = master

    if rm_negative:
        final['Dispatch'] = np.where(final['Dispatch'] < 0, 0, final['Dispatch'])

    # Return dataset
    if return_all:
        return final
    else:
        return final[['Time', 'DUID', 'Dispatch']]


def _clean_duplicates(table, value_col="Dispatch"):
    """Clean duplicate data of Time and DUID from dataframe"""
    if any(table.duplicated(subset=['Time', 'DUID'])):
        print("Duplicate Timestamped DUIDs found. Updating dataset for duplicates.")
        # Take average values where duplicates differ
        table_clean = table.pivot_table(index=["Time", "DUID"], values=value_col, aggfunc=np.mean)
        table_clean = table_clean.reset_index()

        # Remove duplicates where Time and DUID match
        table_clean = table_clean.drop_duplicates(subset=["Time", "DUID"])
        return table_clean
    else:
        return table


def _check_interventions(table):
    """Check and account for intervention flag for data from DISPATCHLOAD MMS table. Not required for DISPATCH_UNIT_SCADA table"""
    # Split table into intervals where intervention has occurred or not
    timestamps_w_intervtn = list(table[table["INTERVENTION"] == 1]["Time"].unique())

    if timestamps_w_intervtn:
        print("Intervention periods found. Updating dataset for interventions.")
        data_unchanged = table[~table["Time"].isin(timestamps_w_intervtn)]
        data_intervtn_updated = table[(table["Time"].isin(timestamps_w_intervtn)) & (table["INTERVENTION"] == 1)]

        # Updates table removing intervention == 0 datapoints for intervals where intervention has occurred
        updated_table = pd.concat([data_unchanged, data_intervtn_updated], ignore_index=True)
        updated_table.sort_values(by=["Time", "DUID"], inplace=True)
        return updated_table.reset_index(drop=True)
    else:
        return table


def download_pricesetter_files(cache, start_time, end_time):
    """Download NEM Price Setter files from MMS table.
    First caches raw XML files as JSON and then reads and returns data in the form of pandas.DataFrame.
    Processed data only considers the marginal generator for the Energy market.

    For further explaination on NEMPriceSetting refer to: https://aemo.com.au/-/media/files/electricity/nem/it-systems-and-change/nemde-queue/nemde_queue_users_guide.pdf?la=en

    Parameters
    ----------
    cache : str
        Raw data location in local directory
    start_time : str
        Start Time Period in format 'yyyy/mm/dd HH:MM'
    end_time : str
        End Time Period in format 'yyyy/mm/dd HH:MM'

    Returns
    -------
    pandas.DataFrame

        ============  ========  ================================================================================================
        Columns:      Type:     Description:
        PeriodID      datetime  The NEM market dispatch interval.
        RegionID      str       The NEM market region.
        Price         float     The market price for dispatch interval.
        Unit          str       A DUID who contributes to setting the price (in most cases).
        BandNo        int       Trade band number of the unit's contribution to price setting.
        Increase      float     A marginal increase (in MW) in the unit band for a 1MW increase in energy demand for the region.
        RRNBandPrice  float     Unit Band price as referred to the RRN
        BandCost      float     Amount in $/h (Increase column multiplied by RRNBandPrice)
        ============  ========  ================================================================================================

    """
    # Check inputs
    cache = hp._check_cache(cache)
    start_time = hp._validate_and_convert_date(start_time, "start_time")
    end_time = hp._validate_and_convert_date(end_time, "end_time")

    # Adjust data collection date range
    collect_sdt = datetime(start_time.year, start_time.month, start_time.day)
    if (end_time.hour == 0) & (end_time.minute == 0):
        collect_edt = datetime(end_time.year, end_time.month, end_time.day)
    else:
        collect_edt = datetime(end_time.year, end_time.month, end_time.day) + timedelta(days=1)

    daterange_list = pd.date_range(collect_sdt, collect_edt)

    # Check if any files in cache already exist within daterange, remove them from new dateranges to create json for new dates only
    xml_files = glob.glob(os.path.join(cache, "NEMED_PS_DAILY_*.json"))
    exist_file_list = [datetime.strptime(existing_files[-15:-5],"%Y-%m-%d") for existing_files in xml_files]
    new_daterange_only = [x for x in daterange_list if x not in exist_file_list]

    # Download & Process Price Setter Files
    logger.info("Processing Price Setter Files...")
    for date in tqdm(new_daterange_only):
        try:
            _populate_xml_into_daily_json(cache, year=date.year, month=date.month, day=date.day)
        except:
            logger.warning("PriceSetter Download for {} failed. Continuing with remaining dates...".format(date))

    # Read cached JSON Price Setter Files
    table = read_json_to_df(cache, start_dt=start_time, end_dt=end_time)
    return table


def _populate_xml_into_daily_json(cache, year, month, day, rm_xml=True):
    """Iterative function to extract all price setter xml files for a single day and combined into a single JSON file"""
    sdate = datetime(year, month, day)
    edate = sdate + timedelta(days=1)
    date_str_list = [datetime.strftime(i,"%Y/%m/%d %H:%M:%S") for i in pd.date_range(sdate, edate, freq='5T', \
        inclusive="right")]

    # Load individual xml files
    dataset = []
    for interval in date_str_list:
        xml_cache_manager = XML(cache)
        overwrite_xmlcachemanager_with_pricesetter_config()

        xml_cache_manager.load_interval(interval)
        d = xml_cache_manager.xml['SolutionAnalysis']['PriceSetting']
        dataset += [d]
    dataset = [item for sublist in dataset for item in sublist]

    # Write JSON daily summary file to cache
    write_file = "NEMED_PS_DAILY_" + sdate.strftime("%Y-%m-%d") + ".json"
    json_path = os.path.join(cache, write_file)
    with open(json_path, 'w') as fp:
        json.dump(dataset, fp)

    if rm_xml:
        # Remove XML files cached by nempy XMLCacheManager
        prev_date = datetime(year, month, day) - timedelta(days=1)
        XML_files = glob.glob(os.path.join(cache, "NEMPriceSetter_{}{}{}*.xml"\
            .format(year,str(month).zfill(2),str(day).zfill(2))))
        XML_files += glob.glob(os.path.join(cache, "NEMPriceSetter_{}{}{}*.xml"\
            .format(prev_date.year,str(prev_date.month).zfill(2),str(prev_date.day).zfill(2))))

        for filepath in XML_files:
            os.remove(filepath)
    return



def download_pricesetters(cache, start_year, start_month, start_day, end_year, end_month, end_day,
                          redownload_xml=False):
    # """LEGACY: Deprecated

    # Downloads price setter from AEMO NEMWEB for each dispatch interval if JSON files do not already exist in cache.
    # Returns this data in a pandas dataframe.

    # Parameters
    # ----------
    # cache : str
    #     Raw data location in local directory
    # start_year : int
    #     Year in format 20XX
    # start_month : int
    #     Month from 1..12
    # start_day : int
    #     Day from 1..31
    # end_year : int
    #     Year in format 20XX
    # end_month : int
    #     Month from 1..12
    # end_day : int
    #     Day from 1..31
    # redownload_xml : bool, optional
    #     Setting to True will force new download of XML files irrespective of existing files in cache, by default False.

    # Returns
    # -------
    # pd.DataFrame
    #     Price Setter dataframe containing columns: [PeriodID, RegionID, Market, Price, DUID, DispatchedMarket, BandNo,
    #     Increase, RRNBandPrice, BandCost]
    # """

    # if not redownload_xml:
    #     # Check if JSON files already exist in cache for downloaded data daterange.
    #     start = datetime(year=start_year, month=start_month, day=start_day) - timedelta(days=1)
    #     if end_month == 12:
    #         end_month = 0
    #         end_year += 1
    #     end = datetime(year=end_year, month=end_month, day=end_day)
    #     download_date = start

    #     JSON_files = glob.glob(os.path.join(cache, "*.json"))

    #     while download_date <= end:
    #         searchfor = str(download_date.year) + str(download_date.month).zfill(2) + str(download_date.day).zfill(2)

    #         if not any([item.__contains__(searchfor) for item in JSON_files]):
    #             print("No existing JSON found for date {}".format(download_date))
    #             redownload_xml = True
    #             break
    #         download_date += timedelta(days=1)

    # # Download PriceSetter XML if not found in cache
    # if redownload_xml:
    #     print("Redownloading XML data")
    #     download_pricesetters_xml(cache , start_year, start_month, start_day, end_year, end_month, end_day)

    # print("Reading JSON to pandas Dataframe")
    # start_date_str = str(start_year) + "/" + str(start_month).zfill(2) + "/" + str(start_day).zfill(2)
    # end_date_str = str(end_year) + "/" + str(end_month).zfill(2) + "/" + str(end_day).zfill(2)
    # table = read_json_to_df(cache, start_date_str, end_date_str)
    # return table
    raise Exception("DEPRECATED in this version of NEMED. Refer to documentation: https://nemed.readthedocs.io/en/latest/")




"""Core user interfacing module"""
from . import process as nd
from . helper_functions import helpers as hp
from datetime import datetime as dt, timedelta
import pandas as pd

def get_total_emissions(start_time, end_time, cache, filter_regions=None, by=None, generation_sent_out=True, assume_energy_ramp=True, return_pivot=False):
    """Retrieve (Aggregated) Regional Emissions data for total emissions (absolute and emissions intensity), as well as sent-out
    energy generation for a defined period and time-resolution (e.g. hour, day, month)

    Parameters
    ----------
    start_time : str
        Start Time Period in format 'yyyy/mm/dd HH:MM:SS'
    end_time : str
        End Time Period in format 'yyyy/mm/dd HH:MM:SS'
    cache : str
        Raw data location in local directory
    filter_regions : list(str)
        NEM regions to filter for while retrieving the data, as a list, by default None to collect all region data
    by : str, one of ['interval', 'hour', 'day', 'month', 'year']
        The time-resolution of output data to aggregate to, by default None to return unaggregated 5-minute time resolution
    generation_sent_out : bool
        Considers 'sent_out' generation (auxilary loads) as opposed to 'as generated' in calculations, by default True
    assume_ramp : bool
        Uses a linear ramp between dispatch scada points as opposed to a stepped function, by default True
    return_pivot : bool
        Changes the structure of the returned dataframe to a pivot with column hierarchy as Data Metric then Region, by default False

    Returns
    -------
    pandas.DataFrame
        Data is returned as formatted if `return_pivot` = False:
        
        ===============  ========  ===================================================================================================================
        Columns:         Type:     Description:
        TimeBeginning    datetime  Timestamp for start of interval or aggregation period. Only returned if `by` parameter is set.
        TimeEnding       datetime  Timestamp for end of interval or aggregation period.
        Region           str       The NEM region corresponding to data. 'NEM' field reflects all regions and is returned if `filter_regions` is None. 
        Energy           float     The total (sent-out if `generation_sent_out` is True) energy for the corresponding region and time.
        Total_Emissions  float     The total emissions for the corresponding region and time.
        Intensity_Index  float     The intensity index as above, considering the total emissions divided by (sent-out) energy.
        ===============  ========  ===================================================================================================================
    """
    # Check if cache folder exists
    hp._check_cache(cache)

    # Get emissions for all units by dispatch interval
    raw_table = nd.get_total_emissions_by_DI_DUID(
        start_time, end_time, cache, filter_regions=filter_regions,
        generation_sent_out=generation_sent_out, assume_energy_ramp=assume_energy_ramp, return_all=True)
    clean_table = raw_table.drop_duplicates(subset=['Time', 'DUID'])

    # Aggregate DUID data to regions
    en_colname = clean_table.columns[clean_table.columns.str.contains('Energy')][0]
    res = clean_table[['Time', 'Region', en_colname, 'Total_Emissions']].groupby(['Time', 'Region'])\
        .sum().reset_index()

    # Create NEM agggregation
    if filter_regions == None:
        nem = res.groupby(['Time']).sum().reset_index()
        nem.insert(1,'Region','NEM')
        res = pd.concat([res,nem],ignore_index=True)

    # Aggregate data to `by`
    if by != None:
        aggregate = nd.aggregate_data_by(data=res, by=by)
    else:
        aggregate = res.rename(columns={'Time': 'TimeEnding'})

    # Calculate Intensity Index considering weighted sum of emissions / energy
    aggregate['Intensity_Index'] = aggregate['Total_Emissions'] / aggregate[en_colname]
    aggregate['Intensity_Index'] = aggregate['Intensity_Index'].fillna(0.0)

    # Return as pivot table
    if return_pivot:
        aggregate = aggregate.pivot(index="TimeEnding",
                                    columns="Region",
                                    values=[en_colname, "Total_Emissions", "Intensity_Index"])
    else:
        aggregate = aggregate.sort_values(['TimeEnding','Region']).reset_index(drop=True)
    
    return aggregate


def get_marginal_emissions(start_time, end_time, cache):
    """Retrieves the marginal emissions intensity for each dispatch interval and region. This factor being the weighted
    sum of the generators contributing to price-setting. Although not necessarily common, there may be times where
    multiple technology types contribute to the marginal emissions - note however that the 'DUID' and 'CO2E_ENERGY_SOURCE'
    returned will reflect only the plant which makes the greatest contribution towards price-setting.

    Parameters
    ----------
    cache : str
        Raw data location in local directory
    start_time : str
        Start Time Period in format 'yyyy/mm/dd HH:MM'
    end_time : str
        End Time Period in format 'yyyy/mm/dd HH:MM'

    Returns
    -------
    pandas.DataFrame
        Data is returned as:

        ==================  ========  ==================================================================================================
        Columns:            Type:     Description:
        Time                datetime  Timestamp reported as end of dispatch interval.
        Region              str       The NEM region corresponding to the marginal emitter data. 
        Intensity_Index     float     The intensity index [tCO2e/MWh] (as by weighted contributions) of the price-setting generators.
        DUID                str       Unit identifier of the generator with the largest contribution on the margin for that Time-Region.
        CO2E_ENERGY_SOURCE  str       Unit energy source with the largest contribution on the margin for that Time-Region.
        ==================  ========  ==================================================================================================

    """
    # Check if cache folder exists
    hp._check_cache(cache)

    result = nd.get_marginal_emitter(cache, start_time, end_time)

    return result



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['nemed', 'nemed.data', 'nemed.helper_functions']

package_data = \
{'': ['*'], 'nemed.data': ['plant_auxiliary/*', 'plant_technology/*']}

install_requires = \
['datetime',
 'joblib>=1.2.0,<2.0.0',
 'nemosis',
 'nempy',
 'pandas>=1.2,<2.0',
 'pathlib',
 'plotly',
 'requests',
 'tqdm',
 'xmltodict']

setup_kwargs = {
    'name': 'nemed',
    'version': '0.3.1',
    'description': 'NEM Emissions Data tool',
    'long_description': '# NEMED\n\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Documentation Status](https://readthedocs.org/projects/nemed/badge/?version=latest)](https://nemed.readthedocs.io/en/latest/?badge=latest)\n\nNEMED[^1], or NEM Emissions Data, is a python package to retrieve and process historical emissions data of the National Electricity Market (NEM), reproduced by datasets published by the Australian Energy Market Operator (AEMO).\n\n[^1]: Not to be confused with *"Nemed", "Nimeth"* of the [Irish legend](https://en.wikipedia.org/wiki/Nemed), who was the leader of the third group of people to settle in Ireland.\n\n## Installation\n```bash\npip install nemed\n```\n\n## Introduction\n\nThis tool is designed to allow users to retrieve historical NEM regional emissions data, either total or marginal emissions, for any 5-minute dispatch interval or aggregations thereof. Total emissions data produced by NEMED is given as both absolute total emissions (tCO2-e) and as an emissions intensity index (tCO2-e/MWh). Marginal emissions data reflects the price setter of a particular region, yielding an emissions intensity index (tCO2-e/MWh) corresponding to a particular plant.\nAlthough data is published by AEMO via the [Carbon Dioxide Equivalent Intensity Index (CDEII) Procedure](https://www.aemo.com.au/energy-systems/electricity/national-electricity-market-nem/market-operations/settlements-and-payments/settlements/carbon-dioxide-equivalent-intensity-index) this only reflects a daily summary by region of total and average emissions.\n\n### How does NEMED calculate emissions?\nTotal Emissions are computed by considering 5-minute dispatch interval data for each generator in the NEM for the respective regions, along with their CO2-equivalent emissions factors per unit (generator) level. A detailed method of the process to produce results for: total emissions(tCO2-e) and the corresponidng emisssions intensities can be found [here](https://nemed.readthedocs.io/en/latest/method.html). The tool is able to provide these metrics on a dispatch interval basis, or aggregated to hourly, daily, monthly or yearly measures. For more advanced users, the emissions associated with each generator and hence that generator\'s contribution to total regional emissions can be extracted.\n\nMarginal Emissions are computed by identifying the marginally dispatched generators from AEMO\'s Price Setter files, mapping emissions intensity metrics mentioned above and computing marginal emissions intensity (tCO2-e/MWh).\n\n### How accurate is NEMED?\nA series of [benchmark results](https://nemed.readthedocs.io/en/latest/examples/cdeii_benchmark.html) for total emissions shows a comparison between AEMO\'s daily CDEII reported emissions figures and NEMED\'s emissions figures which have been aggregated from a 5-minute dispatch-interval resolution to a daily basis.   \n\nThe [example](https://nemed.readthedocs.io/en/latest/examples/cdeii_benchmark.html) includes a region by region comparison for each metric, while an overview of the historical NEM Emissions Intensity produced using NEMED is shown here.\n![NEM Emissions Intensity](./docs/source/examples/charts_benchmark/intensity_NEM.png)\n\n## Usage\n\n### Examples\nExamples can be found in [NEMED\'s documentation](https://nemed.readthedocs.io/en/latest/examples/total_emissions.html).\n\n### Possible Use Cases\nSome example use cases of data produced from this tool include:\n- Analysis of historical emissions between NEM regions, generation technologies contributions to them and assessing the difference between total and marginal emissions.\n- Using emissions intensities traces (total and marginal) from NEMED in counter-factual optimisation models; studying the influence of shadow-pricing carbon or imposing carbon constraints.\n- Considering the emissions assosciated with grid-energy consumption for residential/C&I consumers, or in counterfactual studies of hypothetical EV usage or H2 electrolyser operation. \n\n## Contributing\nInterested in contributing? Check out the [contributing guidelines](CONTRIBUTING.md), which also includes steps to install `NEMED` for development.\n\nPlease note that this project is released with a [Code of Conduct](CONDUCT.md). By contributing to this project, you agree to abide by its terms.\n\n## License\n`NEMED` was created by Declan Heim and Shayan Naderi. It is licensed under the terms of the `BSD 3-Clause license`.\n\n## Credits\nThis package was created using the [`UNSW CEEM template`](https://github.com/UNSW-CEEM/ceem-python-template). It also adopts functionality from sister tools including [`NEMOSIS`](https://github.com/UNSW-CEEM/NEMOSIS) and [`NEMPY`](https://github.com/UNSW-CEEM/nempy).',
    'author': 'Declan Heim',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



# read version from installed package
#from importlib.metadata import version
from .nemed import *
#__version__ = version("nemed")
import logging
import sys

logging.getLogger(__name__).addHandler(logging.NullHandler())
logging.basicConfig(
    stream=sys.stdout, level=logging.INFO, format="%(levelname)s: %(message)s"
)